from abc import ABC, abstractmethod
from typing import Any, Type, Callable

from tentacule.i_worker_process import IWorkerProcess


class IProcessPool(ABC):
    worker_class: Type[IWorkerProcess]
    workers: int

    terminate_timeout: int
    result_timeout: int

    @abstractmethod
    def start(self):
        pass

    @abstractmethod
    def close(self, force: bool):
        pass

    @abstractmethod
    def new_task(self, task: Callable, *args, **kwargs) -> str:
        pass

    @abstractmethod
    def get_result(self, task_id: str, timeout: int = 30) -> Any:
        pass



from tests.depency import subtract


def simple_function():
    return 5


def test_simple_function(pool):
    task_id = pool.new_task(simple_function)

    assert pool.get_result(task_id) == 5


def add(a: int, b: int):
    return a + b


def test_simple_function_with_args(pool):
    task_id = pool.new_task(add, 6, 7)

    assert pool.get_result(task_id) == 13


def test_simple_function_with_kwargs(pool):
    task_id = pool.new_task(add, a=5, b=5)

    assert pool.get_result(task_id) == 10


def default_a(a: int = 0):
    return a


def test_simple_function_with_default(pool):
    task_id = pool.new_task(default_a)
    assert pool.get_result(task_id) == 0

    task_id = pool.new_task(default_a, 80)
    assert pool.get_result(task_id) == 80


def recurse(a: int = 0):
    if a == 10:
        return a
    return recurse(a + 1)


def test_recursive(pool):
    task_id = pool.new_task(recurse)
    assert pool.get_result(task_id) == 10


def dependency_subtract(a: int, b: int):
    return subtract(a, b)


def test_function_with_dependency(pool):
    task_id = pool.new_task(dependency_subtract, 7, 2)
    assert pool.get_result(task_id) == 5



import time
from multiprocessing import Queue, Event
from queue import Empty
from threading import Thread
from typing import Any, Type, List, Callable, Optional

import dill

from tentacule.i_process_pool import IProcessPool
from tentacule.i_worker_process import IWorkerProcess
from tentacule.utils import terminate_process_with_timeout, generate_unique_id
from tentacule.worker_process import SimpleWorkerProcess


class ProcessPool(IProcessPool):
    def __init__(self, workers: int, worker_class: Type[IWorkerProcess] = SimpleWorkerProcess):
        self.workers = workers
        self.worker_class = worker_class

        self.terminate_timeout = 5
        self._pool: List[IWorkerProcess] = []
        self._stop_pool = False

        self._task_queue = Queue()
        self._result_queue = Queue()

        self.result_timeout = 15
        self._results = {}
        self._result_events = {}
        self._result_thread: Optional[Thread] = Thread(target=self._watch_for_result, daemon=True)

    def start(self):
        self._pool = [
            SimpleWorkerProcess(self._task_queue, self._result_queue)
            for i in range(self.workers)
        ]

        self._result_thread.start()
        _ = [p.start() for p in self._pool]

    def close(self, force: bool = False):
        if force:
            self._stop_pool = True
            _ = [p.native_process.kill() for p in self._pool]
            self._result_thread.join()
            return

        for process in self._pool:
            terminate_process_with_timeout(process.native_process, self.terminate_timeout)

    def new_task(self, task: Callable, *args, **kwargs) -> str:
        task_id = generate_unique_id()
        self._task_queue.put((task_id, dill.dumps(task), args, kwargs))
        self._result_events[task_id] = Event()

        return task_id

    def get_result(self, task_id: str, timeout: int = 30) -> Any:
        try:
            self._result_events[task_id].wait(timeout)
            return self._results[task_id][0]
        finally:
            self._result_events.pop(task_id)

    def _rebalance(self):
        self._pool = [p for p in self._pool if p.native_process.is_alive()]
        if self.workers == len(self._pool):
            return

        for i in range(self.workers - len(self._pool)):
            process = SimpleWorkerProcess(self._task_queue, self._result_queue)
            self._pool.append(process)

            process.start()

    def _watch_for_result(self):
        while not self._stop_pool:
            try:
                task_id, result = self._result_queue.get(timeout=self.result_timeout)

                self._result_events[task_id].set()
                self._results[task_id] = (result, time.monotonic())

                self._rebalance()
                self._results = {
                    k: v for k, v in self._results.items()
                    if time.monotonic() - v[1] < self.result_timeout
                }
            except Empty:
                pass  # It's okay if the queue was empty, just retry to get



from abc import ABC, abstractmethod
from multiprocessing import Process, Queue


class IWorkerProcess(ABC):
    native_process: Process

    task_queue: Queue
    result_queue: Queue

    @abstractmethod
    def start(self):
        pass



import os
import configparser
import subprocess
import sys
import importlib
from contextlib import contextmanager

from setuptools import setup


@contextmanager
def substituted_config():
    with open('setup.cfg', 'r') as cfg:
        template_cfg = cfg.read()

    if '\n' in template_cfg:
        template_cfg = template_cfg.replace('\r', '')
    else:
        template_cfg = template_cfg.replace('\r', '\n')

    try:
        from envsubst import envsubst
        substituted_config_text = envsubst(template_cfg)
        with open('setup.cfg', 'w') as cfg:
            cfg.write(substituted_config_text)

        yield template_cfg, substituted_config_text
    finally:
        with open('setup.cfg', 'w') as cfg:
            cfg.write(template_cfg)


def raw_install(*packages):
    subprocess.run([sys.executable, '-m', 'pip', 'install', '-U', *packages])
    for p in packages:
        if '=' in p:
            p = p.split('=')[0]
        globals()[p] = importlib.import_module(p)


if __name__ == '__main__':
    raw_install('pip', 'setuptools', 'wheel', 'envsubst', 'pytest', 'nox')

    with substituted_config() as (template, substituted):
        configs = configparser.ConfigParser()
        configs.read_string(substituted)
        requirements = configs['options']['install_requires'].split('\n')[1:]
        # test_requirements = configs['options']['install_test_requires'].split('\n')[1:]

        if len(sys.argv) == 2 and sys.argv[1] == 'install':
            raw_install(*requirements)
            # gitlab_install(personal_access_token, *test_requirements)
        else:
            setup(
                install_requires=requirements
            )



import uuid
from multiprocessing import Process


def generate_unique_id() -> str:
    return uuid.uuid4().hex


def terminate_process_with_timeout(process: Process, timeout: int):
    try:
        process.terminate()
        process.join(timeout)
    except TimeoutError:
        process.kill()
